/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "keymap.h"
#include "send.h"           /* for mutt_edit_file(), in curs_lib.c? */
#include "rfc1524.h"

#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>

/* 
 * Currently, this only works for send mode, as it assumes that the 
 * BODY->filename actually contains the information.  I'm not sure
 * we want to deal with editing attachments we've already received,
 * so this should be ok.
 *
 * Returns 1 if editor found, 0 if not (useful to tell calling menu to
 * redraw)
 */
int mutt_edit_attachment (BODY *a, int opt)
{
  char type[STRING];
  char command[STRING];
  rfc1524_entry *entry = rfc1524_new_entry ();

  snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);
  if (rfc1524_mailcap_lookup (a,type, entry, M_EDIT))
  {
    if (entry->editcommand)
    {
      char newfile[_POSIX_PATH_MAX] = "";

      strfcpy (command, entry->editcommand, sizeof (command));
      if (rfc1524_expand_filename (entry->nametemplate, a->filename, newfile, sizeof (newfile)))
      {
	dprint(1,(debugfile,"oldfile: %s\t newfile: %s\n",a->filename,newfile));
	if (!mutt_rename_file (a->filename, newfile))
	{
	  if (!mutt_yesorno ("Can't match nametemplate, continue?", 1))
	    return 0;
	}
	else
	{
	  safe_free ((void **) &a->filename);
	  a->filename = safe_strdup(newfile);
	}
      }

      if (rfc1524_expand_command (a, a->filename, type, command, sizeof (command)))
      {
	/* For now, editing requires a file, no piping */
	mutt_error("Mailcap Edit entry requires %%s");
      }
      else
      {
	mutt_endwin (NULL);
	mutt_system (command);
	wait (NULL);
      }
    }
  }
  else if (a->type == TYPETEXT)
  {
    /* On text, default to editor */
    mutt_edit_file (strcmp ("builtin", Editor) == 0 ? Visual : Editor, 
		    a->filename);
  }
  else
  {
    rfc1524_free_entry (&entry);
    mutt_error("No mailcap edit entry for %s",type);
    return 0;
  }

  rfc1524_free_entry (&entry);
  return 1;
}

int mutt_is_autoview (char *type)
{
  LIST *t = AutoViewList;
  int i;

  while (t)
  {
    i = strlen (t->data) - 1;
    if ((i > 0 && t->data[i-1] == '/' && t->data[i] == '*' && 
	  strncasecmp (type, t->data, i) == 0) || strcasecmp (type, t->data) == 0)
      return 1;
    t = t->next;
  }

  return 0;
}

/* Ok, this currently assumes send state, as the BODY->filename
 * actually contains the file.
 *
 * This routine will attempt to display the attachment.  It will
 * first look if the type is viewable without mailcap.  If it
 * requires a mailcap, it will display it, otherwise it will use the built-in
 * mutt_body_handler() to display the message.
 *
 * There is currently a problem when displaying multipart messages, as
 * we need to call mutt_parse_mime_message() on the BODY, but that function
 * currently accepts HEADER, which we don't have.  We might get around this
 * by calling parse_messageRFC822() or parse_multipart() directly.
 *
 * This should handle all four cases, ie:
 *   piped input, paged output
 *   piped input, terminal output
 *   piped
 */
int mutt_view_attachment (BODY *a, int opt)
{
   int viewtext = FALSE;
   char tempfile[_POSIX_PATH_MAX];
   
   viewtext = !mutt_needs_mailcap (a);

   if (!viewtext)
   {
     rfc1524_entry *entry = rfc1524_new_entry ();
     char type[STRING];
     char command[_POSIX_PATH_MAX+STRING];
     char descrip[STRING];
     int piped = FALSE;
     FILE *tfp,*ifp,*fpout,*fpin;

     snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);

     viewtext = !rfc1524_mailcap_lookup (a, type, entry, 0);

     if (!viewtext)
     {
       /* Ok, we got an entry, view it */
       char newfile[_POSIX_PATH_MAX] = "";
       if (rfc1524_expand_filename (entry->nametemplate, a->filename, newfile, sizeof (newfile)))
       {
	 if (!mutt_rename_file (a->filename,newfile))
	 {
	   if (mutt_yesorno ("Can't match nametemplate, continue?", 1) != M_YES)
	     return 0;
	 }
	 else
	 {
	   safe_free ((void **) &a->filename);
	   a->filename = safe_strdup (newfile);
	 }
       }

       if (entry->command)
       strfcpy (command, entry->command, sizeof (command));
       else
       {
	 mutt_error ("MIME type not defined.  Cannot view attachment.");
	 rfc1524_free_entry (&entry);
	 return 0;
       }

       piped = rfc1524_expand_command (a, a->filename, type, command, sizeof (command));
       if (entry->copiousoutput)
       {
	 pid_t thepid;

	 mutt_mktemp (tempfile);
	 if ((tfp = safe_fopen (tempfile,"w")) == NULL)
	 {
	   mutt_perror ("fopen");
	   rfc1524_free_entry (&entry);
	   return 0;
	 }

	 if (piped)
	 {
	   if ((ifp = fopen (a->filename,"r")) == NULL)
	   {
	     mutt_perror ("fopen");
	     rfc1524_free_entry (&entry);
	     fclose (tfp);
	     mutt_unlink (tempfile);
	     return 0;
	   }

	   mutt_endwin (NULL);
	   thepid = mutt_create_filter (command, &fpout, &fpin, NULL);
	   mutt_copy_stream (ifp, fpout);
	   mutt_copy_stream (fpin, tfp);
	   mutt_wait_filter (thepid);
	   fclose (fpout);
	   fclose (fpin);
	   fclose (ifp);
	   fclose (tfp);
	 }
	 else
	 {
	   mutt_endwin (NULL);
	   thepid = mutt_create_filter (command, NULL, &fpin, NULL);
	   mutt_copy_stream (fpin, tfp);
	   mutt_wait_filter (thepid);
	   fclose (tfp);
	   fclose (fpin);
	 }

	 if (a->description)
	   snprintf (descrip, sizeof(descrip),
		     "---Command: %-20.20s Description: %s",
		     command,a->description);
	 else
	   snprintf (descrip, sizeof (descrip),
		     "---Command: %-30.30s Attachment: %s/%s", command,
		     TYPE (a->type), a->subtype);
	 mutt_do_pager (descrip, tempfile, 0);
       }
       else
       {
	 /* interactive program */
	 if (piped)
	 {
	   pid_t thepid;

	   if ((ifp = fopen (a->filename,"r")) == NULL)
	   {
	     mutt_perror ("fopen");
	     rfc1524_free_entry (&entry);
	     return 0;
	   }
	   mutt_endwin (NULL);
	   thepid = mutt_create_filter (command, &fpout, NULL, NULL);
	   mutt_copy_stream (ifp, fpout);
	   mutt_wait_filter (thepid);
	   fclose (fpout);
	   fclose (ifp);
	 }
	 else
	 {
	   mutt_endwin (NULL);
	   mutt_system (command);
	 }
       }
     }
     rfc1524_free_entry (&entry);
   }

   if (viewtext)
   {
     /* Use mutt built-in handler */
     STATE s;
     char descrip[STRING];
     struct stat finfo;
     unsigned int save_encoding;
     
     memset (&s, 0, sizeof (s));

     mutt_mktemp (tempfile);
     if ((s.fpout = safe_fopen (tempfile,"w")) == NULL)
     {
       mutt_perror ("fopen");
       return 0;
     }

     /* Have to make copy since mutt_do_pager() deletes the file */
     if ((s.fpin = safe_fopen (a->filename, "r")) == NULL)
     {
       mutt_perror("fopen");
       fclose(s.fpout);
       mutt_unlink(tempfile);
       return 0;
     }

     stat (a->filename, &finfo);
     a->length = finfo.st_size;

     save_encoding = a->encoding;
     a->encoding = ENC8BIT;

     if(a->type == TYPEMESSAGE && a->subtype && 
	(!strcasecmp(a->subtype,"rfc822") || !strcasecmp(a->subtype, "news")))
       a->parts = mutt_parse_messageRFC822(s.fpin, a);
     
     if (a->description)
       strfcpy (descrip, a->description, sizeof (descrip));
     else
       snprintf (descrip, sizeof (descrip),
                 "---Attachment: %s/%s", TYPE (a->type), a->subtype);
     s.displaying = 1;
     mutt_body_handler (a, &s);
     fclose (s.fpout);
     fclose (s.fpin);
     mutt_do_pager (descrip, tempfile, a->type == TYPEMESSAGE && strcasecmp ("rfc822", a->subtype) == 0);
     a->length = 0;
     a->encoding = save_encoding;
     if(a->parts)
       mutt_free_body(&a->parts);
   }
   return 1;
}

/* returns 1 on success, 0 on error */
int mutt_pipe_attachment (FILE *fp, BODY *b, const char *path, char *outfile)
{
  STATE o;
  pid_t thepid;

  memset (&o, 0, sizeof (STATE));

  if (outfile)
    if ((o.fpout = fopen (outfile,"w")) == NULL)
    {
      mutt_perror ("fopen");
      return 0;
    }

  mutt_endwin (NULL);

  if (fp)
  {
    /* recv case */

    STATE s;

    memset (&s, 0, sizeof (STATE));

    if (outfile)
      thepid = mutt_create_filter (path, &s.fpout, &o.fpin, NULL);
    else
      thepid = mutt_create_filter (path, &s.fpout, NULL, NULL);

    s.fpin = fp;
    mutt_decode_attachment (b, &s);
    fclose (s.fpout);
  }
  else
  {
    /* send case */

    FILE *ifp, *ofp;

    if ((ifp = fopen (b->filename, "r")) == NULL)
    {
      mutt_perror ("fopen");
      fclose (o.fpout);
      return 0;
    }

    if (outfile)
      thepid = mutt_create_filter (path, &ofp, &o.fpin, NULL);
    else
      thepid = mutt_create_filter (path, &ofp, NULL, NULL);

    mutt_copy_stream (ifp, ofp);
    fclose (ofp);
    fclose (ifp);
  }

  if (outfile)
  {
    mutt_copy_stream (o.fpin, o.fpout);
    fclose (o.fpin);
    fclose (o.fpout);
  }

  if (mutt_wait_filter (thepid) != 0 || option (OPTWAITKEY))
    ci_any_key_to_continue (NULL);
  return 1;
}

/* returns 0 on success, -1 on error */
int mutt_save_attachment (FILE *fp, BODY *m, const char *path)
{
  if (access (path, F_OK) == 0 && mutt_yesorno ("File exists, overwrite?", 0) != 1)
    return (-1);

  if (fp)
  {
    /* In recv mode, extract from folder and decode */

    STATE s;
  
    memset (&s, 0, sizeof (s));
    if ((s.fpout = fopen (path, "w")) == NULL)
    {
      mutt_perror ("fopen");
      return (-1);
    }
    fseek ((s.fpin = fp), m->offset, 0);
    mutt_decode_attachment (m, &s);

    if (fclose (s.fpout) != 0)
    {
      mutt_perror ("fclose");
      return (-1);
    }
  }
  else
  {
    /* In send mode, just copy file */

    FILE *ofp, *nfp;

    if ((ofp = fopen (m->filename,"r")) == NULL)
    {
      mutt_perror ("fopen");
      return (-1);
    }

    if ((nfp = fopen (path, "w")) == NULL)
    {
      mutt_perror ("fopen");
      fclose (ofp);
      return (-1);
    }

    mutt_copy_stream (ofp, nfp);
    fclose (ofp);
    fclose (nfp);
  }

  return 0;
}

/* Ok, the difference between send and receive:
 * recv: BODY->filename is a suggested name, and Context|HEADER points
 *       to the attachment in mailbox which is encooded
 * send: BODY->filename points to the un-encoded file which contains the 
 *       attachment
 */

int mutt_print_attachment (FILE *fp, BODY *a)
{
  char type[STRING];
  rfc1524_entry *entry = rfc1524_new_entry();

  snprintf (type, sizeof (type), "%s/%s", TYPE (a->type), a->subtype);

  if (rfc1524_mailcap_lookup (a, type, entry, M_PRINT)) 
  {
    /* Ok, we got an entry, view it */

    char newfile[_POSIX_PATH_MAX] = "";
    char command[_POSIX_PATH_MAX+STRING];
    int piped = FALSE;
    FILE *ifp,*fpout;

    if (rfc1524_expand_filename (entry->nametemplate, a->filename, newfile, sizeof (newfile)))
    {
      if (!fp)
      {
	/* only attempt file move in send mode */

	if (!mutt_rename_file (a->filename, newfile))
	{
	  if (mutt_yesorno ("Can't match nametemplate, continue?", 1) != M_YES)
	  {
	    rfc1524_free_entry (&entry);
	    return 0;
	  }
	  strfcpy (newfile, a->filename, sizeof (newfile));
	}
	else
	{
	  safe_free ((void **)&a->filename);
	  a->filename = safe_strdup (newfile);
	}
      }
    }

    /* in recv mode, save file to newfile first */
    if (fp)
      mutt_save_attachment (fp, a, newfile);

    strfcpy (command, entry->printcommand, sizeof (command));
    piped = rfc1524_expand_command (a, newfile, type, command, sizeof (command));

    mutt_endwin (NULL);

    /* interactive program */
    if (piped)
    {
      pid_t thepid;

      if ((ifp = fopen (newfile,"r")) == NULL)
      {
	mutt_perror ("fopen");
	rfc1524_free_entry (&entry);
	return 0;
      }

      thepid = mutt_create_filter (command, &fpout, NULL, NULL);
      mutt_copy_stream (ifp, fpout);
      fclose (fpout);
      fclose (ifp);
      mutt_wait_filter (thepid);
      ci_any_key_to_continue (NULL);
    }
    else
      mutt_system (command);

    if (fp)
      mutt_unlink (newfile);

    rfc1524_free_entry (&entry);
    return 1;
  } 

  rfc1524_free_entry (&entry);

  if (!strcasecmp ("text/plain",a->subtype) || 
      !strcasecmp ("application/postscript", a->subtype))
  {
    return (mutt_pipe_attachment (fp, a, PrintCmd, NULL));
  }
  else
  {
    mutt_error ("I don't know how to print that!");
    return 0;
  }
}
