/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"

#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>

/* HP-UX and ConvexOS don't have this macro */
#ifndef S_ISLNK
#define S_ISLNK(x) (((x) & S_IFMT) == S_IFLNK ? 1 : 0)
#endif

struct folder_file
{
  mode_t mode;
  time_t mtime;
  off_t size;
  char *name;
  char *desc;
};

/*
 * Global variables for this package.
 */
static struct folder_file *Entries = NULL;
static short LastEntry = 0; /* number of real entries */
static short MaxEntry = 0;  /* max entry */
static char LastDir[_POSIX_PATH_MAX] = "";

/* Frees up the memory allocated for the local-global variables.  */
static void destroy_global (void)
{
  int c;

  for (c = 0; c < LastEntry; c++)
  {
    safe_free ((void **) &Entries[c].name);
    safe_free ((void **) &Entries[c].desc);
  }
  safe_free ((void **) &Entries);
}

static int browser_compare_subject (const void *a, const void *b)
{
  struct folder_file *pa = (struct folder_file *) a;
  struct folder_file *pb = (struct folder_file *) b;

  int r = strcmp (pa->name, pb->name);

  return ((BrowserSort & SORT_REVERSE) ? -r : r);
}

static int browser_compare_date (const void *a, const void *b)
{
  struct folder_file *pa = (struct folder_file *) a;
  struct folder_file *pb = (struct folder_file *) b;

  int r = pa->mtime - pb->mtime;

  return ((BrowserSort & SORT_REVERSE) ? -r : r);
}

static int browser_compare_size (const void *a, const void *b)
{
  struct folder_file *pa = (struct folder_file *) a;
  struct folder_file *pb = (struct folder_file *) b;

  int r = pa->size - pb->size;

  return ((BrowserSort & SORT_REVERSE) ? -r : r);
}

static void browser_sort (void)
{
  int (*f) (const void *, const void *);

  switch (BrowserSort & SORT_MASK)
  {
    case SORT_ORDER:
      return;
    case SORT_DATE:
      f = browser_compare_date;
      break;
    case SORT_SIZE:
      f = browser_compare_size;
      break;
    case SORT_SUBJECT:
    default:
      f = browser_compare_subject;
      break;
  }
  qsort (Entries, LastEntry, sizeof (struct folder_file), f);
}

static int link_is_dir (const char *path)
{
  struct stat st;

  if (stat (path, &st) == 0)
    return (S_ISDIR (st.st_mode));
  else
    return (-1);
}

static void add_folder (const char *name, const struct stat *s, int new)
{
  char buffer[_POSIX_PATH_MAX + SHORT_STRING];
  struct passwd *pw;
  struct group *gr;
  char date[16];
  char *fmt;
  time_t tnow;

  pw = getpwuid (s->st_uid);
  gr = getgrgid (s->st_gid);
  
  tnow = time (NULL);
  fmt = tnow - s->st_mtime < 31536000 ? "%b %d %H:%M" : "%b %d  %Y";
  strftime (date, sizeof (date), fmt, localtime (&s->st_mtime));
  
  snprintf (buffer, sizeof (buffer), "%c %c%c%c%c%c%c%c%c%c%c %2d %-8.8s %-8.8s %8ld %s %s%s",
 	    new ? 'N' : ' ',
 	    S_ISDIR(s->st_mode) ? 'd' : (S_ISLNK(s->st_mode) ? 'l' : '-'),
 	    
 	    (s->st_mode & S_IRUSR) != 0 ? 'r': '-',
 	    (s->st_mode & S_IWUSR) != 0 ? 'w' : '-',
 	    (s->st_mode & S_ISUID) != 0 ? 's' : (s->st_mode & S_IXUSR) != 0 ? 'x': '-',
 	    
 	    (s->st_mode & S_IRGRP) != 0 ? 'r' : '-',
 	    (s->st_mode & S_IWGRP) != 0 ? 'w' : '-',
 	    (s->st_mode & S_ISGID) != 0 ? 's' : (s->st_mode & S_IXGRP) != 0 ? 'x': '-',
 	    
 	    (s->st_mode & S_IROTH) != 0 ? 'r' : '-',
 	    (s->st_mode & S_IWOTH) != 0 ? 'w' : '-',
 	    (s->st_mode & S_ISVTX) != 0 ? 't' : (s->st_mode & S_IXOTH) != 0 ? 'x': '-',
 	    s->st_nlink,
 	    pw ? pw->pw_name : "(null)",
 	    gr ? gr->gr_name : "(null)",
 	    (long) s->st_size,
 	    date,
 	    name,
 	    S_ISLNK (s->st_mode) ? "@" : (S_ISDIR (s->st_mode) ? "/" : ((s->st_mode & S_IXUSR) != 0 ? "*" : "")));
  
  if (LastEntry == MaxEntry)
  {
    /* need to allocate more space */
    safe_realloc ((void **) &Entries, sizeof (struct folder_file) * (MaxEntry += 256));
  }

  Entries[LastEntry].mode = s->st_mode;
  Entries[LastEntry].mtime = s->st_mtime;
  Entries[LastEntry].size = s->st_size;
  Entries[LastEntry].name = safe_strdup (name);
  Entries[LastEntry].desc = safe_strdup (buffer);

  LastEntry++;
}

static int examine_directory (const char *d, const char *prefix)
{
  struct stat s;
  DIR *dp;
  struct dirent *de;
  char buffer[_POSIX_PATH_MAX + SHORT_STRING];
  BUFFY *tmp;

  if (stat (d, &s) == -1)
  {
    mutt_perror (d);
    return (-1);
  }

  if (!S_ISDIR (s.st_mode))
  {
    mutt_error ("%s is not a directory", d);
    return (-1);
  }

  mutt_buffy_check ();

  if ((dp = opendir (d)) == NULL)
  {
    mutt_perror (d);
    return (-1);
  }

  LastEntry = 0;
  MaxEntry = 256;
  Entries = (struct folder_file *) safe_malloc (sizeof (struct folder_file) * MaxEntry);

  while ((de = readdir (dp)) != NULL)
  {
    if (strcmp (de->d_name, ".") == 0)
      continue;    /* we don't need . */
    
    if (prefix && *prefix && strncmp (prefix, de->d_name, strlen (prefix)) != 0)
      continue;
    if (regexec ((regex_t *) Mask.rx, de->d_name, 0, NULL, 0) != 0)
      continue;

    snprintf (buffer, sizeof (buffer), "%s/%s", d, de->d_name);
    if (lstat (buffer, &s) == -1)
      continue;
    
    if ((! S_ISREG (s.st_mode)) && (! S_ISDIR (s.st_mode)) &&
	(! S_ISLNK (s.st_mode)))
      continue;
    
    tmp = Incoming;
    while (tmp && strcmp(buffer, tmp->path))
      tmp = tmp->next;
    add_folder(de->d_name, &s, (tmp) ? tmp->new : 0);
  }
  closedir (dp);  
  browser_sort ();
  return 0;
}

static int examine_mailboxes (void)
{
  struct stat s;
  char buffer[LONG_STRING];
  BUFFY *tmp = Incoming;

  if (!Incoming)
    return (-1);
  mutt_buffy_check();

  LastEntry = 0;
  MaxEntry = 256;
  Entries = (struct folder_file *) safe_malloc (sizeof (struct folder_file) * MaxEntry);

  do
  {
    if (lstat (tmp->path, &s) == -1)
      continue;

    if ((! S_ISREG (s.st_mode)) && (! S_ISDIR (s.st_mode)) &&
	(! S_ISLNK (s.st_mode)))
      continue;
    
    strncpy (buffer, tmp->path, sizeof (buffer));
    mutt_pretty_mailbox (buffer);

    add_folder (buffer, &s, tmp->new);
  }
  while ((tmp = tmp->next));
  browser_sort ();
  return 0;
}

int select_file_search (regex_t *re, int n)
{
  return (regexec (re, Entries[n].name, 0, NULL, 0));
}

void folder_entry (char *s, size_t slen, MUTTMENU *m, int num)
{
  strfcpy (s, Entries[num].desc, slen);
}

static void
init_menu (MUTTMENU *menu, char *title, size_t titlelen, int buffy)
{
  char path[_POSIX_PATH_MAX];

  menu->current = 0;
  menu->top = 0;
  menu->max = LastEntry;
  if (buffy)
    snprintf (title, titlelen, "Mailboxes [%d]", mutt_buffy_check ());
  else
  {
    strfcpy (path, LastDir, sizeof (path));
    mutt_pretty_mailbox (path);
    snprintf (title, titlelen, "Directory [%s]", path);
  }
  menu->redraw = REDRAW_FULL;
}

void mutt_select_file (char *f, size_t flen, int buffy)
{
  char buf[STRING];
  char prompt[STRING];
  char prefix[_POSIX_PATH_MAX] = "";
  char helpstr[SHORT_STRING];
  char title[STRING];
  MUTTMENU *menu;
  struct stat st;
  int i, killPrefix = 0;
  
  if (*f)
  {
    mutt_expand_path (f, flen);
    for (i = strlen (f) - 1; i > 0 && f[i] != '/' ; i--);
    if (i > 0)
    {
      if (f[0] == '/')
      {
	if (i > sizeof (LastDir) - 1) i = sizeof (LastDir) - 1;
	strncpy (LastDir, f, i);
	LastDir[i] = 0;
      }
      else
      {
	getcwd (LastDir, sizeof (LastDir));
	strcat (LastDir, "/");
	strncat (LastDir, f, i);
      }
    }
    else
    {
      if (f[0] == '/')
	strcpy (LastDir, "/");
      else
	getcwd (LastDir, sizeof (LastDir));
    }

    if (i <= 0 && f[0] != '/')
      strfcpy (prefix, f, sizeof (prefix));
    else
      strfcpy (prefix, f + i + 1, sizeof (prefix));
    killPrefix = 1;
  }
  else if (!LastDir[0])
    strfcpy (LastDir, Maildir, sizeof (LastDir));

  *f = 0;

  if (buffy)
  {
    if (examine_mailboxes () == -1)
      return;
  }
  else if (examine_directory (LastDir, prefix) == -1)
      return;

  menu = mutt_new_menu ();
  menu->menu = MENU_FOLDER;
  menu->make_entry = folder_entry;
  menu->search = select_file_search;
  menu->title = title;

  helpstr[0] = 0;
  mutt_make_help (buf, sizeof (buf), "Exit  ", MENU_FOLDER, OP_EXIT);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Chdir  ", MENU_FOLDER, OP_CHANGE_DIRECTORY);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Mask  ", MENU_FOLDER, OP_ENTER_MASK);
  strcat (helpstr, buf);
  mutt_make_help (buf, sizeof (buf), "Help", MENU_FOLDER, OP_HELP);
  strcat (helpstr, buf);
  menu->help = helpstr;

  init_menu (menu, title, sizeof (title), buffy);

  FOREVER
  {
    switch (i = mutt_menuLoop (menu))
    {
      case OP_GENERIC_SELECT_ENTRY:

	if (S_ISDIR (Entries[menu->current].mode) ||
	    (S_ISLNK (Entries[menu->current].mode) && link_is_dir (Entries[menu->current].name)))
	{
	  /* make sure this isn't a MH or maildir mailbox */
	  if (buffy)
	  {
	    strncpy (buf, Entries[menu->current].name, sizeof(buf));
	    mutt_expand_path (buf, sizeof (buf));
	  }
	  else
	    snprintf (buf, sizeof (buf), "%s/%s", LastDir, Entries[menu->current].name);

	  if (mx_get_magic (buf) <= 0)
	  {
	    char OldLastDir[_POSIX_PATH_MAX];

	    /* save the old directory */
	    strfcpy (OldLastDir, LastDir, sizeof (OldLastDir));

	    if (strcmp (Entries[menu->current].name, "..") == 0)
	    {
	      if (strcmp ("..", LastDir + strlen (LastDir) - 2) == 0)
		strcat (LastDir, "/..");
	      else
	      {
		char *p = strrchr (LastDir + 1, '/');

		if (p)
		  *p = 0;
		else
		{
		  if (LastDir[0] == '/')
		    LastDir[1] = 0;
		  else
		    strcat (LastDir, "/..");
		}
	      }
	    }
	    else if (buffy)
	    {
	      sprintf (LastDir, "%s", Entries[menu->current].name);
	      mutt_expand_path(LastDir, sizeof(LastDir));
	    }
	    else
	      sprintf (LastDir + strlen (LastDir), "/%s", Entries[menu->current].name);

	    destroy_global ();
	    if (killPrefix)
	    {
	      prefix[0] = 0;
	      killPrefix = 0;
	    }
	    buffy = 0;
	    if (examine_directory (LastDir, prefix) == -1)
	    {
	      /* try to restore the old values */
	      strfcpy (LastDir, OldLastDir, sizeof (LastDir));
	      if (examine_directory (LastDir, prefix) == -1)
	      {
		strfcpy (LastDir, Homedir, sizeof (LastDir));
		return;
	      }
	    }
	    init_menu (menu, title, sizeof (title), buffy);
	    break;
	  }
	}

	if (buffy)
	{
	  strncpy(f, Entries[menu->current].name, flen);
	  mutt_expand_path(f, flen);
	}
	else
	snprintf (f, flen, "%s/%s", LastDir, Entries[menu->current].name);

	/* Fall through to OP_EXIT */

      case OP_EXIT:

	destroy_global ();
	mutt_menuDestroy (&menu);
	return;

      case OP_CHANGE_DIRECTORY:

	buffy = 0;	  
	strfcpy (buf, LastDir, sizeof (buf));
	if (ci_get_field ("Chdir to: ", buf, sizeof (buf), M_FILE) == 0 &&
	    buf[0])
	{
	  mutt_expand_path (buf, sizeof (buf));
	  if (stat (buf, &st) == 0)
	  {
	    if (S_ISDIR (st.st_mode))
	    {
	      strfcpy (LastDir, buf, sizeof (LastDir));
	      destroy_global ();
	      if (examine_directory (LastDir, prefix) == 0)
		init_menu (menu, title, sizeof (title), buffy);
	      else
	      {
		mutt_error ("Error scanning directory.");
		destroy_global ();
		mutt_menuDestroy (&menu);
		return;
	      }
	    }
	    else
	      mutt_error ("%s is not a directory.", buf);
	  }
	  else
	    mutt_perror (buf);
	}
	break;
	
      case OP_ENTER_MASK:
	
	buffy = 0;
	if (!Mask.pattern)
	  exit (42);
	else
	  strfcpy (buf, Mask.pattern, sizeof (buf));
	i = 0;
	do
	{
	  snprintf (prompt, sizeof (prompt), "%sFile mask: ",
					      (i ? "No matching files. " : ""));
	  if (ci_get_field (prompt, buf, sizeof (buf), 0) == 0)
	  {
	    regex_t *rx = (regex_t *) safe_malloc (sizeof (regex_t));
	    int err;

	    i = 0;
	    /* assume that the user wants to see everything */
	    if (!buf[0])
	      strfcpy (buf, ".", sizeof (buf));

	    if ((err = regcomp (rx, buf, REG_EXTENDED)) != 0)
	    {
	      regerror (err, rx, buf, sizeof (buf));
	      regfree (rx);
	      safe_free ((void **) &rx);
	      mutt_error ("%s", buf);
	    }
	    else
	    {
	      safe_free ((void **) &Mask.pattern);
	      regfree ((regex_t *) Mask.rx);
	      safe_free ((void **) &Mask.rx);
	      Mask.pattern = safe_strdup (buf);
	      Mask.rx = rx;

	      destroy_global ();
	      if (examine_directory (LastDir, NULL) == 0)
		init_menu (menu, title, sizeof (title), buffy);
	      else
	      {
		mutt_error ("Error scanning directory.");
		mutt_menuDestroy (&menu);
		return;
	      }
	      killPrefix = 0;
	    }
	  }

	  if (LastEntry == 0)
	    i = 1;
	} while (i);
	break;

      case OP_SORT:
      case OP_SORT_REVERSE:

	{
	  int reverse = 0;

	  move (LINES - 1, 0);
	  if (i == OP_SORT_REVERSE)
	  {
	    reverse = SORT_REVERSE;
	    addstr ("Reverse ");
	  }
	  addstr ("Sort by (d)ate, (a)lpha, si(z)e or do(n)'t sort? ");
	  clrtoeol ();

	  while ((i = ci_getch ()) != EOF && i != 'a' && i != 'd' && i != 'z'
		 && i != 'n')
	  {
	    if (i == ERR || CI_is_return (i))
	      break;
	    else
	      BEEP ();
	  }

	  if (i != EOF)
	  {
	    switch (i)
	    {
	      case 'a': 
	        BrowserSort = reverse |  SORT_SUBJECT;
		break;
	      case 'd':
	        BrowserSort = reverse |  SORT_DATE;
		break;
	      case 'z': 
	        BrowserSort = reverse |  SORT_SIZE;
		break;
	      case 'n': 
	        BrowserSort = SORT_ORDER;
		break;
	    }
	    browser_sort ();
	    menu->redraw = REDRAW_FULL;
	  }
	}

	break;

      case OP_CHECK_NEW:

	buffy = 1 - buffy;
	destroy_global ();
	prefix[0] = 0;
	killPrefix = 0;
	if (buffy)
	{
	  if (examine_mailboxes () == -1)
	    return;
	}
	else if (examine_directory (LastDir, prefix) == -1)
	  return;
	init_menu (menu, title, sizeof (title), buffy);
	break;
    }
  }
  /* not reached */
}
