/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_regex.h"
#include "rfc2047.h"

#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <stdlib.h>

/* Reads an arbitrarily long header field, and looks ahead for continuation
 * lines.  ``line'' is dynamically increased if more space is required to fit
 * the whole line.
 */
static char *mutt_read_rfc822_line (FILE *f, char *line, size_t *linelen)
{
  size_t offset = 0, l;
  char buf[LONG_STRING];
  char ch, *pbuf, *q;
  int is_cont = 0; /* continuation line? */
  int is_full;

  if (line)
    *line = 0;

  FOREVER
  {
    if (fgets (buf, sizeof (buf), f) == NULL)
      return (line);

    /* check to see if we got a full line */
    l = strlen (buf);
    is_full = (!l || buf[l - 1] == '\n');

    pbuf = buf;
    if (is_cont)
    {
      /* unfold the continuation line, as described by MIME */
      for (; ISSPACE (*pbuf); pbuf++, l--) /* empty */
	;
      if (*pbuf)
      {
	*--pbuf = ' '; /* convert to a single space */
	l++;
      }
    }
    is_cont = 0;

    /* if we got a full line, remove trailing space */
    if (l && pbuf[l - 1] == '\n')
    {
      for (q = pbuf + l - 1 ; q >= pbuf && ISSPACE (*q) ; q--, l--)
        *q = 0;
    }

    if (offset + l > *linelen)
    {
      /* not enough space to hold the (rest of the) line, grow the buffer */
      if (line)
      {
	*linelen = offset + l;
	safe_realloc ((void **) &line, *linelen + 1);
      }
      else
      {
	*linelen = l;
	line = safe_malloc (*linelen + 1);
      }
    }
    else if (!l && !line)
    {
      /* Avoid a segfault if the first line we read is blank. */
      return (NULL);
    }

    if (l)
    {
      memcpy (line + offset, pbuf, l + 1);
      offset += l;
    }

    /* check to see if we got a full line */
    if (!is_full)
      continue; /* no, read the next part */

    /* a blank line indicates the end of the message header */
    if (! *line)
      return (line);

    /* check to see if the next line is a continuation line */
    ch = fgetc (f);
    ungetc (ch, f);
    if (ch != ' ' && ch != '\t')
      return (line); /* next line is a separate header field or EOH */
    is_cont = 1;
  }
  /* not reached */
}

static LIST *mutt_parse_references (char *s)
{
  LIST *t, *lst = NULL;

  while ((s = strtok (s, " \t")) != NULL)
  {
    /*
     * some mail clients add other garbage besides message-ids, so do a quick
     * check to make sure this looks like a valid message-id
     */
    if (*s == '<')
    {
      t = (LIST *)safe_malloc (sizeof (LIST));
      t->data = safe_strdup (s);
      t->next = lst;
      lst = t;
    }
    s = NULL;
  }

  return (lst);
}

int mutt_check_encoding (const char *c)
{
  if (strncasecmp ("7bit", c, sizeof ("7bit")-1) == 0)
    return (ENC7BIT);
  else if (strncasecmp ("8bit", c, sizeof ("8bit")-1) == 0)
    return (ENC8BIT);
  else if (strncasecmp ("binary", c, sizeof ("binary")-1) == 0)
    return (ENCBINARY);
  else if (strncasecmp ("quoted-printable", c, sizeof ("quoted-printable")-1) == 0)
    return(ENCQUOTEDPRINTABLE);
  else if (strncasecmp("base64", c, sizeof("base64")-1) == 0)
    return(ENCBASE64);
  else
    return(ENCOTHER);
}

static PARAMETER *parse_parameters (const char *s)
{
  PARAMETER *head = 0, *cur = 0, *new;
  char buffer[LONG_STRING];
  const char *p;
  size_t i;

  while (*s)
  {
    if ((p = strchr (s, '=')) == NULL)
    {
      dprint(1, (debugfile, "parse_parameters: malformed parameter: %s\n", s));
      return (head); /* just bail out now */
    }

    i = p - s;

    new = mutt_new_parameter ();

    new->attribute = safe_malloc (i + 1);
    memcpy (new->attribute, s, i);
    new->attribute[i] = 0;

    s = p + 1; /* skip over the = */
    SKIPWS (s);

    if (*s == '"')
    {
      s++;
      for (i=0; *s && *s != '"' && i < sizeof (buffer) - 1; i++, s++)
      {
        if (*s == '\\')
	{
          /* Quote the next character */
          buffer[i] = s[1];
	  if (!*++s)
	    break;
        }
        else
          buffer[i] = *s;
      }
      buffer[i] = 0;
      if (*s)
	s++; /* skip over the " */
    }
    else
    {
      for (i=0; *s && *s != ' ' && *s != ';' && i < sizeof (buffer) - 1; i++, s++)
        buffer[i] = *s;
      buffer[i] = 0;
    }

    new->value = safe_strdup (buffer);

    /* Add this parameter to the list */
    if (head)
    {
      cur->next = new;
      cur = cur->next;
    }
    else
      head = cur = new;

    /* Find the next parameter */
    if (*s == ';')
      s++;
    else
    {
      while (*s && *s != ';')
	s++;
      if (!*s)
	break; /* no more parameters! */
      s++;
    }

    /* Move past any leading whitespace */
    SKIPWS (s);

    /* skip empty parameters */
    while (*s == ';')
    {
      s++;
      SKIPWS (s);
    }
  }    

  return (head);
}

int mutt_check_mime_type (const char *s)
{
  if (strcasecmp ("text", s) == 0)
    return TYPETEXT;
  else if (strcasecmp ("multipart", s) == 0)
    return TYPEMULTIPART;
  else if (strcasecmp ("application", s) == 0)
    return TYPEAPPLICATION;
  else if (strcasecmp ("message", s) == 0)
    return TYPEMESSAGE;
  else if (strcasecmp ("image", s) == 0)
    return TYPEIMAGE;
  else if (strcasecmp ("audio", s) == 0)
    return TYPEAUDIO;
  else if (strcasecmp ("video", s) == 0)
    return TYPEVIDEO;
  else
    return TYPEOTHER;
}

static void parse_content_type (char *s, BODY *ct)
{
  char *pc;
  char buffer[SHORT_STRING];
  short i = 0;

  safe_free((void **)&ct->subtype);
  mutt_free_parameter(&ct->parameter);

  /* First extract any existing parameters */
  if ((pc = strchr(s, ';')) != NULL)
  {
    *pc++ = 0;
    while (*pc && whitespace(*pc)) pc++;
    ct->parameter = parse_parameters(pc);

    /* Some pre-RFC1521 gateways still use the "name=filename" convention */
    if ((pc = mutt_get_parameter("name", ct->parameter)) != 0)
      ct->filename = safe_strdup(pc);
  }
  
  /* Now get the subtype */
  if ((pc = strchr(s, '/')))
  {
    *pc++ = 0;
    while (*pc && !whitespace (*pc) && *pc != ';')
    {
      buffer[i++] = *pc;
      pc++;
    }
    buffer[i] = 0;
    ct->subtype = safe_strdup (buffer);
  }

  /* Finally, get the major type */
  ct->type = mutt_check_mime_type (s);

  if (ct->subtype == NULL)
  {
    /* Some older non-MIME mailers (i.e., mailtool, elm) have a content-type
     * field, so we can attempt to convert the type to BODY here.
     */
    if (ct->type == TYPETEXT)
      ct->subtype = safe_strdup ("plain");
    else if (ct->type == TYPEAUDIO)
      ct->subtype = safe_strdup ("basic");
    else if (ct->type == TYPEMESSAGE)
      ct->subtype = safe_strdup ("rfc822");
    else if (ct->type == TYPEOTHER)
    {
      ct->type = TYPEAPPLICATION;
      snprintf (buffer, sizeof (buffer), "x-%s", s);
      ct->subtype = safe_strdup (buffer);
    }
    else
      ct->subtype = safe_strdup ("x-unknown");
  }
}

/* args:
 *	fp	stream to read from
 *
 *	digest	1 if reading subparts of a multipart/digest, 0
 *		otherwise
 */

BODY *mutt_read_mime_header (FILE *fp, int digest)
{
  BODY *p = mutt_new_body();
  PARAMETER *parms;
  char *c, *line = NULL;
  size_t linelen = 0;
  
  p->hdr_offset = ftell(fp);

  p->encoding = ENC7BIT; /* default from RFC1521 */
  p->type = digest ? TYPEMESSAGE : TYPETEXT;

  while ((line = mutt_read_rfc822_line (fp, line, &linelen)) && *line != 0)
  {
    /* Find the value of the current header */
    if ((c = strchr (line, ':')))
    {
      c++;
      SKIPWS (c);
    }
    else
    {
      dprint (1, (debugfile, "read_mime_header: bogus BODY header: %s\n", line));
      break;
    }

    switch (tolower (line[0]))
    {
      case 'c':

	if (strncasecmp("ontent-type", line+1, sizeof("ontent-type")-1) == 0)
	  parse_content_type(c, p);
	else if (strncasecmp("ontent-transfer-encoding", line+1, sizeof("ontent-transfer-encoding")-1) == 0)
	  p->encoding = mutt_check_encoding(c);
	else if (strncasecmp("ontent-disposition", line+1, sizeof("ontent-disposition")-1) == 0)
	{
	  if (strncasecmp ("attachment", c, sizeof ("attachment") - 1) == 0)
	    p->disposition = DISPATTACH;
	  else
	    p->disposition = DISPINLINE;

	  /* Check to see if a default filename was given */
	  if ((c = strchr (c, ';')) != NULL)
	  {
	    c++;
	    SKIPWS (c);
	    if ((c = mutt_get_parameter ("filename", (parms = parse_parameters (c)))) != 0)
	    {
	      /* free() here because the content-type parsing routine might
	       * have allocated space if a "name=filename" parameter was
	       * specified.
	       */
	      safe_free ((void **) &p->filename);
	      p->filename = safe_strdup (c); 
	    }
	    mutt_free_parameter (&parms);
	  }
	}
	else if (!strncasecmp ("ontent-description", line+1, sizeof("ontent-description")-1))
	{
	  if (!p->description)
	    p->description = safe_strdup (c);
	}
	break;
    }
  }
  p->offset = ftell (fp); /* Mark the start of the real data */
  if (p->type == TYPETEXT && !p->subtype)
    p->subtype = safe_strdup ("plain");
  else if (p->type == TYPEMESSAGE && !p->subtype)
    p->subtype = safe_strdup ("rfc822");

  safe_free ((void **) &line);

  return (p);
}

/* parse a MESSAGE/RFC822 body
 *
 * args:
 *	fp		stream to read from
 *
 *	parent		structure which contains info about the message/rfc822
 *			body part
 *
 * NOTE: this assumes that `parent->length' has been set!
 */

BODY *mutt_parse_messageRFC822 (FILE *fp, BODY *parent)
{
  BODY *msg;

  msg = mutt_read_mime_header (fp, 0);
  msg->length = parent->length - (msg->offset - parent->offset);

  /* if body of this message is empty, we can end up with a negative length */
  if (msg->length < 0)
    msg->length = 0;

  if (msg->type == TYPEMULTIPART)
    msg->parts = mutt_parse_multipart (fp, mutt_get_parameter ("boundary", msg->parameter), msg->offset + msg->length, strcasecmp ("digest", msg->subtype) == 0);
  else if (msg->type == TYPEMESSAGE)
    msg->parts = mutt_parse_messageRFC822 (fp, msg);
  else
    return (msg);
  
  /* try to recover from parsing error */
  if (!msg->parts)
  {
    msg->type = TYPETEXT;
    safe_free ((void **) &msg->subtype);
    msg->subtype = safe_strdup ("plain");
  }

  return (msg);
}

/* parse a multipart structure
 *
 * args:
 *	fp		stream to read from
 *
 *	boundary	body separator
 *
 *	end_off		length of the multipart body (used when the final
 *			boundary is missing to avoid reading too far)
 *
 *	digest		1 if reading a multipart/digest, 0 otherwise
 */

BODY *mutt_parse_multipart (FILE *fp, const char *boundary, long end_off, int digest)
{
  int blen, len, crlf = 0;
  char buffer[LONG_STRING];
  BODY *head = 0, *last = 0, *new = 0;
  int i;
  int final = 0; /* did we see the ending boundary? */

  if (!boundary)
  {
    mutt_error ("multipart message has no boundary parameter!");
    return (NULL);
  }

  blen = strlen (boundary);
  while (ftell (fp) < end_off && fgets (buffer, LONG_STRING, fp) != NULL)
  {
    len = strlen (buffer);

    /* take note of the line ending.  I'm assuming that either all endings
     * will use <CR><LF> or none will.
     */
    if (len > 1 && buffer[len - 2] == '\r')
      crlf = 1;

    if (buffer[0] == '-' && buffer[1] == '-' &&
	strncmp (buffer + 2, boundary, blen) == 0)
    {
      if (last)
      {
        last->length = ftell (fp) - last->offset - len - 1 - crlf;
        if (last->parts && last->parts->length == 0)
          last->parts->length = ftell (fp) - last->parts->offset - len - 1 - crlf;
      }

      /* Remove any trailing whitespace, up to the length of the boundary */
      for (i = len - 1; whitespace (buffer[i]) && i > blen + 2; i--)
        buffer[i] = 0;

      /* Check for the end boundary */
      if (buffer[2 + blen] == '-' &&
	  buffer[3 + blen] == '-' &&
	  buffer[4 + blen] == 0)
      {
	final = 1;
        break;
      }

      new = mutt_read_mime_header (fp, digest);
      if (head)
      {
        last->next = new;
        last = new;
      }
      else
        last = head = new;
    }
  }
  
  /* in case of missing end boundary, set the length to something reasonable */
  if (last && last->length == 0 && !final)
    last->length = end_off - last->offset;

  /* parse recursive MIME parts */
  last = head;
  while (last)
  {
    fseek (fp, last->offset, 0);
    switch (last->type)
    {
      case TYPEMULTIPART:

	last->parts = mutt_parse_multipart (fp, mutt_get_parameter ("boundary", last->parameter), last->offset + last->length, strcasecmp ("digest", last->subtype) == 0);
	break;

      case TYPEMESSAGE:

	if (last->subtype &&
	    (strcasecmp (last->subtype, "rfc822") == 0 ||
	     strcasecmp (last->subtype, "news") == 0))
	  last->parts = mutt_parse_messageRFC822 (fp, last);
	break;
    }

    last = last->next;
  }

  return (head);
}

static struct tz_t
{
  char *tzname;
  unsigned char zhours;
  unsigned char zminutes;
  unsigned char zoccident; /* west of UTC? */
  unsigned char xxx;       /* unused */
}
TimeZones[] =
{
  { "sst",  11,  0, 1, 0 }, /* Samoa */
  { "pst",   8,  0, 1, 0 },
  { "mst",   7,  0, 1, 0 },
  { "pdt",   7,  0, 1, 0 },
  { "cst",   6,  0, 1, 0 },
  { "mdt",   6,  0, 1, 0 },
  { "cdt",   5,  0, 1, 0 },
  { "est",   5,  0, 1, 0 },
  { "ast",   4,  0, 1, 0 }, /* Atlantic */
  { "edt",   4,  0, 1, 0 },
  { "wgt",   3,  0, 1, 0 }, /* Western Greenland */
  { "wgst",  2,  0, 1, 0 }, /* Western Greenland DST */
  { "aat",   1,  0, 1, 0 }, /* Atlantic Africa Time */
  { "egt",   1,  0, 1, 0 }, /* Eastern Greenland */
  { "egst",  0,  0, 0, 0 }, /* Eastern Greenland DST */
  { "gmt",   0,  0, 0, 0 },
  { "utc",   0,  0, 0, 0 },
  { "wat",   0,  0, 0, 0 }, /* West Africa */
  { "wet",   0,  0, 0, 0 }, /* Western Europe */
  { "bst",   1,  0, 0, 0 }, /* British DST */
  { "cat",   1,  0, 0, 0 }, /* Central Africa */
  { "cet",   1,  0, 0, 0 }, /* Central Europe */
  { "met",   1,  0, 0, 0 }, /* this is now officially CET */
  { "west",  1,  0, 0, 0 }, /* Western Europe DST */
  { "cest",  2,  0, 0, 0 }, /* Central Europe DST */
  { "eet",   2,  0, 0, 0 }, /* Eastern Europe */
  { "ist",   2,  0, 0, 0 }, /* Israel */
  { "sat",   2,  0, 0, 0 }, /* South Africa */
  { "ast",   3,  0, 0, 0 }, /* Arabia */
  { "eat",   3,  0, 0, 0 }, /* East Africa */
  { "eest",  3,  0, 0, 0 }, /* Eastern Europe DST */
  { "idt",   3,  0, 0, 0 }, /* Israel DST */
  { "msk",   3,  0, 0, 0 }, /* Moscow */
  { "adt",   4,  0, 0, 0 }, /* Arabia DST */
  { "msd",   4,  0, 0, 0 }, /* Moscow DST */
  { "gst",   4,  0, 0, 0 }, /* Presian Gulf */
  { "smt",   4,  0, 0, 0 }, /* Seychelles */
  { "ist",   5, 30, 0, 0 }, /* India */
  { "ict",   7,  0, 0, 0 }, /* Indochina */
  { "cst",   8,  0, 0, 0 }, /* China */
  { "hkt",   8,  0, 0, 0 }, /* Hong Kong */
  { "wst",   8,  0, 0, 0 }, /* Western Australia */
  { "jst",   9,  0, 0, 0 }, /* Japan */
  { "cst",   9, 30, 0, 0 }, /* Australian Central Standard Time */
  { "kst",  10,  0, 0, 0 }, /* Korea */
  { "nzst", 12,  0, 0, 0 }, /* New Zealand */
  { NULL,    0,  0, 0, 0 }
};

/* parses a date string in RFC822 format:
 *
 * Date: [ weekday , ] day-of-month month year hour:minute:second timezone
 *
 */
static int parse_date (char *s, HEADER *h)
{
  int count = 0;
  char *t;
  int hour, min, sec;
  struct tm tm;
  int i;
  int tz_offset = 0;

  h->zoccident = 0;

  /* kill the day of the week, if it exists. */
  if ((t = strchr (s, ',')) != 0)
  {
    t++;
    SKIPWS (t);
  }
  else
    t = s;

  memset (&tm, 0, sizeof (tm));

  while ((t = strtok (t, " \t")) != NULL)
  {
    switch (count)
    {
      case 0: /* day of the month */
	if (!isdigit (*t))
	  return (-1);
	tm.tm_mday = atoi (t);
	if (tm.tm_mday > 31)
	  return (-1);
	break;

      case 1: /* month of the year */
	if ((i = mutt_check_month (t)) < 0)
	  return (-1);
	tm.tm_mon = i;
	break;

      case 2: /* year */
	tm.tm_year = atoi (t);
	if (tm.tm_year >= 1900)
	  tm.tm_year -= 1900;
	break;

      case 3: /* time of day */
	if (sscanf (t, "%d:%d:%d", &hour, &min, &sec) == 3)
	  ;
	else if (sscanf (t, "%d:%d", &hour, &min) == 2)
	  sec = 0;
	else
	{
	  dprint(1, (debugfile, "parse_date: could not process time format: %s\n", t));
	  return(-1);
	}
	tm.tm_hour = hour;
	tm.tm_min = min;
	tm.tm_sec = sec;
	break;

      case 4: /* timezone */
	if (*t == '+' || *t == '-')
	{
	  if (t[1] && t[2] && t[3] && t[4])
	  {
	    h->zhours = (t[1] - '0') * 10 + (t[2] - '0');
	    h->zminutes = (t[3] - '0') * 10 + (t[4] - '0');

	    if (t[0] == '-')
	      h->zoccident = 1;
	  }
	}
	else
	{
	  struct tz_t *p = TimeZones;
	  int i;

	  for (i = 0; p[i].tzname; i++)
	  {
	    if (!strcasecmp (p[i].tzname, t))
	    {
	      h->zhours = p[i].zhours;
	      h->zminutes = p[i].zminutes;
	      h->zoccident = p[i].zoccident;
	    }
	  }

	  /* ad hoc support for the European MET (now officially CET) TZ */
	  if (strcasecmp (t, "MET") == 0)
	  {
	    if ((t = strtok (NULL, " \t")) != NULL)
	    {
	      if (!strcasecmp (t, "DST"))
		h->zhours++;
	    }
	  }
	}
	tz_offset = h->zhours * 3600 + h->zminutes * 60;
	if (!h->zoccident)
	  tz_offset = -tz_offset;
    }
    count++;
    t = 0;
  }

  if (count < 5)
  {
    dprint(1,(debugfile, "parse_date(): error parsing date format, using received time\n"));
    return (-1);
  }

  h->date_sent = mutt_mktime (&tm) + tz_offset;

  return 0;
}

/* extract the first substring that looks like a message-id */
static char *extract_message_id (const char *s)
{
  const char *p;
  char *r;
  size_t l;

  if ((s = strchr (s, '<')) == NULL || (p = strchr (s, '>')) == NULL)
    return (NULL);
  l = (size_t)(p - s) + 1;
  r = safe_malloc (l + 1);
  memcpy (r, s, l);
  r[l] = 0;
  return (r);
}

void mutt_parse_mime_message (HEADER *cur)
{
  MESSAGE *msg;

  if (cur->content->type != TYPEMESSAGE && cur->content->type != TYPEMULTIPART)
    return; /* nothing to do */

  if ((msg = mx_open_message (Context, cur->msgno)))
  {
    fseek (msg->fp, cur->content->offset, 0);

    if (cur->content->type == TYPEMULTIPART)
    {
      if (!cur->content->parts)
	cur->content->parts = mutt_parse_multipart (msg->fp, mutt_get_parameter ("boundary", cur->content->parameter), cur->content->offset + cur->content->length, strcasecmp ("digest", cur->content->subtype) == 0);
    }
    else
    {
      if (!cur->content->parts)
	cur->content->parts = mutt_parse_messageRFC822 (msg->fp, cur->content);
    }
      

    /* try to recover from parsing error */
    if (!cur->content->parts)
    {
      cur->content->type = TYPETEXT;
      safe_free ((void **) &cur->content->subtype);
      cur->content->subtype = safe_strdup ("plain");
    }
    
    mx_close_message (&msg);
  }
}

/*
 * mutt_read_rfc822_header() -- parses a RFC822 header
 *
 * Args:
 *
 * f		stream to read from
 *
 * hdr		header structure of current message (optional).  If hdr is
 *		NULL, then we are reading a postponed message, or called
 *		from mutt_edit_headers() so we should keep a list of the
 *		user-defined headers.
 *
 * flags	M_STRICT_FIELDS: do check for missing newline between header
 *		and body
 *
 */
ENVELOPE *mutt_read_rfc822_header (FILE *f, HEADER *hdr)
{
  ENVELOPE *e = mutt_new_envelope();
  LIST *last = NULL;
  char *line = NULL, *p;
  char in_reply_to[SHORT_STRING] = "";
  long loc;
  int matched;
  int date_found = 0;
  size_t linelen = 0;

  if (hdr)
  {
    hdr->content = mutt_new_body ();

    /* set the defaults from RFC1521 */
    hdr->content->type = TYPETEXT;
    hdr->content->subtype = safe_strdup ("plain");
    hdr->content->encoding = ENC7BIT;
    hdr->content->length = -1;
  }

  loc = ftell (f);
  while ((line = mutt_read_rfc822_line (f, line, &linelen)) != NULL &&
	 *line != 0)
  {
    matched = 0;

    if ((p = strpbrk (line, ": \t")) == NULL || *p != ':')
    {
      char return_path[LONG_STRING] = { 0 };
      time_t t;

      /* some bogus MTAs will quote the original "From " line */
      if (strncmp (">From ", line, 6) == 0)
      {
	loc = ftell (f);
	continue; /* just ignore */
      }
      else if ((t = is_from (line, return_path, sizeof (return_path))))
      {
	/* MH somtimes has the From_ line in the middle of the header! */
	if (!hdr->received)
	  hdr->received = t;
	loc = ftell (f);
	continue;
      }

      fseek (f, loc, 0);
      break; /* end of header */
    }

    p++;
    SKIPWS (p);

    switch (tolower (line[0]))
    {
      case 'a':
	if (strncasecmp (line+1, "pparently-to:", 13) == 0)
	{
	  rfc822_parse_adrlist (&e->to, p, "@");
	  matched = 1;
	}
	else if (strncasecmp (line+1, "pparently-from:", 15) == 0)
	{
	  rfc822_parse_adrlist (&e->from, p, "@");
	  matched = 1;
	}
	break;

      case 'b':
	if (strncasecmp (line+1, "cc:", 3) == 0)
	{
	  rfc822_parse_adrlist (&e->bcc, p, "@");
	  matched = 1;
	}
	break;

      case 'c':
	if (strncasecmp (line+1, "c:", 2) == 0)
	{
	  rfc822_parse_adrlist (&e->cc, p, "@");
	  matched = 1;
	}
	else if (strncasecmp (line + 1, "ontent-", 7) == 0)
	{
	  if (strncasecmp (line+8, "type:", 5) == 0)
	  {
	    if (hdr)
	      parse_content_type (p, hdr->content);
	    matched = 1;
	  }
	  else if (strncasecmp (line+8, "transfer-encoding:", 18) == 0)
	  {
	    if (hdr)
	      hdr->content->encoding = mutt_check_encoding (p);
	    matched = 1;
	  }
	  else if (strncasecmp (line+8, "length:", 7) == 0)
	  {
	    if (hdr)
	      hdr->content->length = atoi (p);
	    matched = 1;
	  }
	  else if (strncasecmp (line+8, "description:", 12) == 0)
	  {
	    if (hdr)
	      hdr->content->description = safe_strdup (p);
	    matched = 1;
	  }
	}
	break;

      case 'd':
	if (strncasecmp ("ate:", line+1, 4) == 0)
	{
	  if (hdr && parse_date (p, hdr) == 0)
	    date_found = 1;
	}
	break;

      case 'f':
	if (strncasecmp (line+1, "rom:", 4) == 0)
	{
	  rfc822_parse_adrlist (&e->from, p, "@");
	  matched = 1;
	}
	break;

      case 'i':
	if (strncasecmp (line+1, "n-reply-to:", 11) == 0)
	{
	  if (hdr) strfcpy (in_reply_to, p, sizeof (in_reply_to));
	}
	break;

      case 'l':
	if (strncasecmp (line+1, "ines:", 5) == 0)
	{
	  if (hdr) hdr->lines = atoi(p);
	  matched = 1;
	}
	break;

      case 'm':
	if (strncasecmp (line+1, "ime-version:", 12) == 0)
	{
	  if (hdr) hdr->mime = 1;
	  matched = 1;
	}
	else if (strncasecmp (line+1, "essage-id:", 10) == 0)
	{
	  safe_free ((void **)&e->message_id);
	  e->message_id = extract_message_id (p);
	  matched = 1;
	}
	break;

      case 'r':
	if (strncasecmp (line+1, "eferences:", 10) == 0)
	{
	  mutt_free_list (&e->references);
	  e->references = mutt_parse_references (p);
	  matched = 1;
	}
	else if (strncasecmp (line+1, "eply-to:", 8) == 0)
	{
	  rfc822_parse_adrlist (&e->reply_to, p, "@");
	  matched = 1;
	}
	else if (strncasecmp (line+1, "eturn-path:", 11) == 0)
	{
	  if (hdr) rfc822_parse_adrlist (&e->return_path, p, "@");
	  matched = 1;
	}
	break;

      case 's':
	if (strncasecmp (line+1, "ubject:", 7) == 0)
	{
	  if (!e->subject) e->subject = safe_strdup (p);
	  matched = 1;
	}
	else if (strncasecmp (line+1, "ender:", 6) == 0)
	{
	  rfc822_parse_adrlist (&e->sender, p, "@");
	  matched = 1;
	}
	else if (strncasecmp (line+1, "tatus:", 6) == 0)
	{
	  if (hdr)
	  {
	    while (*p)
	    {
	      switch(*p)
	      {
		case 'r':
		  hdr->replied = 1;
		  break;
		case 'O':
		  if (option (OPTMARKOLD))
		    hdr->old = 1;
		  break;
		case 'R':
		  hdr->read = 1;
		  break;
	      }
	      p++;
	    }
	  }
	  matched = 1;
	}
	break;

      case 't':
	if (strncasecmp (line+1, "o:", 2) == 0)
	{
	  rfc822_parse_adrlist (&e->to, p, "@");
	  matched = 1;
	}
	break;

      case 'x':
	if (strncasecmp (line+1, "-status:", 8) == 0)
	{
	  if (hdr)
	  {
	    while (*p)
	    {
	      switch (*p)
	      {
		case 'A':
		  hdr->replied = 1;
		  break;
		case 'D':
		  hdr->deleted = 1;
		  break;
		case 'F':
		  hdr->flagged = 1;
		  break;
		default:
		  break;
	      }
	      p++;
	    }
	  }
	  matched = 1;
	}
	break;

      default:
	break;
    }

    /* if hdr==NULL, then we are using this to parse either a postponed
     * message, or a outgoing message (edit_hdrs), so we want to keep
     * track of the user-defined headers
     */
    if (!matched && !hdr)
    {
      if (last)
      {
	last->next = mutt_new_list ();
	last = last->next;
      }
      else
	last = e->userhdrs = mutt_new_list ();
      last->data = safe_strdup (line);
    }

    loc = ftell (f);
  }

  safe_free ((void **) &line);

  if (hdr)
  {
    hdr->content->hdr_offset = hdr->offset;
    hdr->content->offset = ftell (f);

    /* Check for missing or invalid date.  If this is the case, convert the
     * date the message was received to UTC and use that value.
     */
    if (!date_found)
    {
      dprint(1,(debugfile,"read_rfc822_header(): no date found, using received time from msg separator\n"));
      hdr->date_sent = hdr->received;
    }

    /* if an in-reply-to was given, check to see if it is in the references
     * list already.  if not, add it so we can do a better job of threading.
     */
    if (in_reply_to[0] && (p = extract_message_id (in_reply_to)) != NULL)
    {
      if (!e->references ||
	  (e->references && strcmp (e->references->data, p) != 0))
      {
	LIST *tmp = mutt_new_list ();

	tmp->data = p;
	tmp->next = e->references;
	e->references = tmp;
      }
      else
	safe_free ((void **) &p);
    }

    /* do RFC2047 decoding */
    rfc2047_decode_adrlist (e->from);
    rfc2047_decode_adrlist (e->to);
    rfc2047_decode_adrlist (e->cc);
    rfc2047_decode_adrlist (e->reply_to);

    if (e->subject)
    {
      regmatch_t pmatch[1];
      
      rfc2047_decode (e->subject, e->subject, strlen (e->subject) + 1);
      if (regexec ((regex_t *) ReplyRegexp.rx, e->subject, 1, pmatch, 0) == 0)
	e->real_subj = e->subject + pmatch[0].rm_eo;
      else
	e->real_subj = e->subject;
    }
  }

  return (e);
}

void mutt_parse_adrlist (ADDRESS **p, const char *s, const char *host)
{
  const char *q;

  /* check for a simple whitespace separated list of addresses */
  if ((q = strpbrk (s, "\"<>():;,\\")) == NULL)
  {
    char tmp[HUGE_STRING];
    char *r;

    strfcpy (tmp, s, sizeof (tmp));
    r = tmp;
    while ((r = strtok (r, " \t")) != NULL)
    {
      rfc822_parse_adrlist (p, r, host);
      r = NULL;
    }
  }
  else
    rfc822_parse_adrlist (p, s, host);
}
