/*
 * Copyright (C) 1996,1997 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"

#include <string.h>
#include <ctype.h>
#include <unistd.h>

/*
 * %b = number of incoming folders with unread messages
 * %d = number of deleted messages [option]
 * %f = full mailbox path
 * %F = number of flagged messages [option]
 * %h = hostname
 * %l = length of mailbox (in bytes) [option]
 * %m = total number of messages
 * %M = number of messages shown (virutal message count when limiting) [option]
 * %n = number of new messages [option]
 * %p = number of postponed messages [option]
 * %r = readonly/wontwrite/changed flag
 * %s = current sorting method
 * %t = # of tagged messages [option]
 * %v = Mutt version
 *
 * %>X = right justify and pad with "X"
 * %|X = pad with "X" to end of line
 *
 * optional fields:
 *    %?%M/?
 */

void mutt_status_line (char *buf, size_t buflen, char *p)
{
  char *cp;
  char *wptr = buf;
  char tmp[STRING];
  char prefix[LONG_STRING];
  char fmt[SHORT_STRING];
  char optstring[SHORT_STRING];
  char ch;
  char padchar;
  int wlen = 0;
  int optional;
  int num_post;
  int len;

  buflen--; /* save room for the trailing \0 */
  while (*p && wlen < buflen)
  {
    if (*p == '%')
    {
      p++;

      if (*p == '?')
      {
	p++;
	optional = 1;
      }
      else
	optional = 0;

      cp = prefix;
      while (*p && (isdigit(*p) || *p == '.' || *p == '-'))
	*cp++ = *p++;
      *cp = 0;

      ch = *p;

      cp = optstring;
      if (optional)
      {
	if (*p) p++;
	while (*p && *p != '?')
	  *cp++ = *p++;
      }
      *cp = 0;

      tmp[0] = 0;
      switch (ch)
      {
	case '>': /* right justify rest of line */

	  p++;
	  padchar = *p++;
	  mutt_status_line (tmp, sizeof (tmp), p);
	  len = COLS - wlen - strlen (tmp);
	  while (len > 0)
	  {
	    *wptr++ = padchar;
	    wlen++;
	    len--;
	  }
	  break;

	case '|': /* pad to end of line */

	  p++;
	  padchar = *p;
	  if (buflen > COLS)
	    buflen = COLS;
	  while (wlen < buflen)
	  {
	    *wptr++ = padchar;
	    wlen++;
	  }
	  break;

        case 'b':

	  if (!optional || mutt_buffy_check ())
	  {
	    snprintf (fmt, sizeof (fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof (tmp), fmt, mutt_buffy_check (), optstring);
	  }
	  break;

	case 'd':

	  if (!optional || (Context && Context->deleted))
	  {
	    snprintf (fmt, sizeof (fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof (tmp), fmt,
		      Context ? Context->deleted : 0, optstring);
	  }
	  break;

	case 'h':

	  snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
	  snprintf (tmp, sizeof (tmp), fmt, Hostname);
	  break;

	case 'f':

	  snprintf (fmt, sizeof(fmt), "%%%ss", prefix);
	  if (Context && Context->path)
	  {
	    strfcpy (prefix, Context->path, sizeof (prefix));
	    mutt_pretty_mailbox (prefix);
	  }
	  else
	    strfcpy (prefix, "(no mailbox)", sizeof (prefix));
	  snprintf (tmp, sizeof (tmp), fmt, prefix);
	  break;

	case 'F':

	  if (!optional || (Context && Context->flagged))
	  {
	    snprintf (fmt, sizeof (fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof (tmp), fmt,
		      Context ? Context->flagged : 0, optstring);
	  }
	  break;

	case 'l':

	  if (!optional || (Context && Context->size))
	  {
	    char size[5];

	    snprintf (fmt, sizeof (fmt), "%%%ss%%s", prefix);
	    mutt_pretty_size (size, sizeof (size), Context ? Context->size : 0);
	    snprintf (tmp, sizeof (tmp), fmt, size, optstring);
	  }
	  break;

	case 'm':

	  if (!optional || (Context && Context->msgcount))
	  {
	    snprintf (fmt, sizeof (fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof (tmp), fmt, Context ? Context->msgcount : 0, optstring);
	  }
	  break;

	case 'M':

	  if (!optional || (Context && (Context->vcount != Context->msgcount)))
	  {
	    snprintf (fmt, sizeof(fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof(tmp), fmt,
		      Context ? Context->vcount : 0, optstring);
	  }
	  break;

	case 'n':

	  if (!optional || (Context && Context->new))
	  {
	    snprintf (fmt, sizeof (fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof (tmp), fmt,
		      Context ? Context->new : 0, optstring);
	  }
	  break;

	case 'p':

	  num_post = mutt_num_postponed ();
	  if (!optional || num_post)
	  {
	    snprintf (fmt, sizeof (fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof (tmp), fmt, num_post, optstring);
	  }
	  break;

	case 'r':

	  if (Context)
	    tmp[0] = (Context->readonly || Context->dontwrite) ? StChars[2] :
		      (Context->changed || Context->deleted) ? StChars[1] : StChars[0];
	  else
	    tmp[0] = StChars[0];
	  tmp[1] = 0;
	  break;

	case 's':

	  snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
	  snprintf (prefix, sizeof (prefix), "%s%s",
		    (Sort & SORT_REVERSE) ? "reverse-" : "",
		    mutt_getnamebyvalue ((Sort & SORT_MASK), SortMethods));
	  snprintf (tmp, sizeof (tmp), fmt, prefix);
	  break;

	case 't':

	  if (!optional || (Context && Context->tagged))
	  {
	    snprintf (fmt, sizeof (fmt), "%%%sd%%s", prefix);
	    snprintf (tmp, sizeof (tmp), fmt,
		      Context ? Context->tagged : 0, optstring);
	  }
	  break;

	case 'v':

	  snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
	  snprintf (tmp, sizeof (tmp), fmt, MuttVersion);
	  break;

	case 0:

	  *buf = 0;
	  return;

	default:

	  snprintf (tmp, sizeof (tmp), "%%%s%c", prefix, ch);
	  break;
      }
      len = strlen (tmp);
      if (len + wlen > buflen)
      {
	if ((len = buflen - wlen) < 0)
	  len = 0;
      }
      memcpy (wptr, tmp, len);
      wptr += len;
      wlen += len;
    }
    else
    {
      *wptr++ = *p;
      wlen++;
    }
    p++;
  }
  *wptr = 0;
}
